import rules
import urlman
from django.conf import settings
from django.db import models, transaction
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from django_scopes import ScopedManager, scopes_disabled

from servala.core import rules as perms
from servala.core.models.mixins import ServalaModelMixin
from servala.core.odoo import CLIENT


class Organization(ServalaModelMixin, models.Model):
    name = models.CharField(max_length=100, verbose_name=_("Name"))
    # The namespace is generated as "org-{id}" in accordance with RFC 1035 Label Names.
    # It is nullable as we need to write to the database in order to read the ID, but should
    # not be null in practical use.
    namespace = models.CharField(
        max_length=20,
        verbose_name=_("Kubernetes Namespace"),
        null=True,
        unique=True,
        help_text=_("This namespace will be used for all Kubernetes resources."),
    )

    billing_entity = models.ForeignKey(
        to="BillingEntity",
        on_delete=models.PROTECT,
        related_name="organizations",
        verbose_name=_("Billing entity"),
        null=True,  # TODO: billing entity should be required
    )
    origin = models.ForeignKey(
        to="OrganizationOrigin",
        on_delete=models.PROTECT,
        related_name="organizations",
        verbose_name=_("Origin"),
    )

    members = models.ManyToManyField(
        to="core.User",
        through="OrganizationMembership",
        related_name="organizations",
        verbose_name=_("Members"),
    )

    class urls(urlman.Urls):
        base = "/org/{self.slug}/"
        details = "{base}details/"
        services = "{base}services/"
        instances = "{base}instances/"

    @cached_property
    def slug(self):
        return f"{slugify(self.name)}-{self.id}"

    def get_absolute_url(self):
        return self.urls.base

    def set_owner(self, user):
        with scopes_disabled():
            OrganizationMembership.objects.filter(user=user, organization=self).delete()
            OrganizationMembership.objects.create(
                user=user, organization=self, role=OrganizationRole.OWNER
            )

    @classmethod
    def create_organization(cls, instance, owner):
        try:
            instance.origin
        except Exception:
            instance.origin = OrganizationOrigin.objects.get(
                pk=settings.SERVALA_DEFAULT_ORIGIN
            )
        instance.save()
        instance.set_owner(owner)
        return instance

    class Meta:
        verbose_name = _("Organization")
        verbose_name_plural = _("Organizations")
        rules_permissions = {
            "view": rules.is_staff | perms.is_organization_member,
            "change": rules.is_staff | perms.is_organization_admin,
            "delete": rules.is_staff | perms.is_organization_owner,
            "add": rules.is_authenticated,
        }

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if not self.namespace:
            # Set namespace after initial save to ensure we have an ID
            self.namespace = f"org-{self.pk}"
            self.save(update_fields=["namespace"])


class BillingEntity(ServalaModelMixin, models.Model):
    """
    Every organization has a billing entity, though billing entities
    may be shared across different organizations.
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    erp_reference = models.CharField(
        max_length=100, blank=True, verbose_name=_("ERP reference")
    )

    # Odoo IDs are nullable for creation, should never be null in practice
    # The company ID points at a record of type res.partner with company_type=company
    # The invoice ID points at a record of type res.partner with company_type=person,
    # type=invoic, parent_id=company_id (the invoice address).
    odoo_company_id = models.IntegerField(null=True)
    odoo_invoice_id = models.IntegerField(null=True)

    class Meta:
        verbose_name = _("Billing entity")
        verbose_name_plural = _("Billing entities")

    def __str__(self):
        return self.name

    @classmethod
    @transaction.atomic
    def create_from_data(cls, name, odoo_data):
        """
        Creates a BillingEntity and corresponding Odoo records.

        This method creates a `res.partner` record in Odoo with `company_type='company'`
        for the main company, and another `res.partner` record with `company_type='person'`
        and `type='invoice'` (linked via `parent_id` to the first record) for the
        invoice address. The IDs of these Odoo records are stored in the BillingEntity.

        Args:
            odoo_data (dict): A dictionary containing the data for creating
                              the BillingEntity and Odoo records.

        Expected keys in `odoo_data`:
            - `invoice_street` (str): Street for the invoice address.
            - `invoice_street2` (str): Second line of street address for the invoice address.
            - `invoice_city` (str): City for the invoice address.
            - `invoice_zip` (str): ZIP/Postal code for the invoice address.
            - `invoice_country` (int): Odoo `res.country` ID for the invoice address country.
            - `invoice_email` (str): Email address for the invoice contact.
            - `invoice_phone` (str): Phone number for the invoice contact.
        """
        instance = cls.objects.create(name=name)
        company_payload = {
            "name": odoo_data.get("company_name", name),
            "company_type": "company",
        }
        company_id = CLIENT.execute("res.partner", "create", [company_payload])
        instance.odoo_company_id = company_id

        invoice_address_payload = {
            "name": name,
            "company_type": "person",
            "type": "invoice",
            "parent_id": company_id,
        }
        invoice_optional_fields = {
            "street": odoo_data.get("invoice_street"),
            "street2": odoo_data.get("invoice_street2"),
            "city": odoo_data.get("invoice_city"),
            "zip": odoo_data.get("invoice_zip"),
            "country_id": odoo_data.get("invoice_country"),
            "email": odoo_data.get("invoice_email"),
        }
        invoice_address_payload.update(
            {k: v for k, v in invoice_optional_fields.items() if v is not None}
        )

        invoice_address_id = CLIENT.execute(
            "res.partner", "create", [invoice_address_payload]
        )
        instance.odoo_invoice_id = invoice_address_id

        instance.save(update_fields=["odoo_company_id", "odoo_invoice_id"])
        return instance

    @classmethod
    @transaction.atomic
    def create_from_id(cls, name, odoo_id):
        parent_data = CLIENT.search_read(
            model="res.partner",
            domain=[["id", "=", odoo_id]],
            fields=["parent_id"],
            limit=1,
        )
        # Data validation: If the data is not as expected, we just return None,
        # rather than raising an exception, for now.
        if not parent_data:
            return
        if not (parent_info := parent_data[0].get("parent_id")):
            return
        if not isinstance(parent_info, (list, tuple)) or not len(parent_info) > 0:
            # parent_info is a tuple of the parent’s ID and name
            return

        instance = cls.objects.create(
            name=name, odoo_invoice_id=odoo_id, odoo_company_id=parent_info[0]
        )
        return instance

    @cached_property
    def odoo_data(self):
        data = {
            "company": None,
            "invoice_address": None,
        }

        company_fields = ["name", "company_type"]
        invoice_address_fields = [
            "name",
            "company_type",
            "type",
            "parent_id",
            "street",
            "street2",
            "city",
            "zip",
            "country_id",
            "email",
        ]

        if self.odoo_company_id:
            company_records = CLIENT.search_read(
                model="res.partner",
                domain=[["id", "=", self.odoo_company_id]],
                fields=company_fields,
                limit=1,
            )
            if company_records:
                data["company"] = company_records[0]

        if self.odoo_invoice_id:
            invoice_address_records = CLIENT.search_read(
                model="res.partner",
                domain=[["id", "=", self.odoo_invoice_id]],
                fields=invoice_address_fields,
                limit=1,
            )
            if invoice_address_records:
                data["invoice_address"] = invoice_address_records[0]

        return data


class OrganizationOrigin(ServalaModelMixin, models.Model):
    """
    Every organization has an origin, though origins may be
    shared across different organizations. The default origin
    (by PK) is configured in the settings as SERVALA_DEFAULT_ORIGIN.
    """

    name = models.CharField(max_length=100, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))

    class Meta:
        verbose_name = _("Organization origin")
        verbose_name_plural = _("Organization origins")

    def __str__(self):
        return self.name


class OrganizationRole(models.TextChoices):
    MEMBER = "member", _("Member")
    ADMIN = "admin", _("Administrator")
    OWNER = "owner", _("Owner")


class OrganizationMembership(ServalaModelMixin, models.Model):
    """Through-model for the many-to-many relationship between organizations and users."""

    user = models.ForeignKey(
        to="core.User",
        on_delete=models.CASCADE,
        related_name="organization_memberships",
        verbose_name=_("User"),
    )
    organization = models.ForeignKey(
        to=Organization,
        on_delete=models.CASCADE,
        related_name="memberships",
        verbose_name=_("Organization"),
    )
    date_joined = models.DateTimeField(
        verbose_name=_("Date joined"), auto_now_add=True, editable=False
    )
    role = models.CharField(
        max_length=20,
        choices=OrganizationRole.choices,
        default=OrganizationRole.MEMBER,
        verbose_name=_("Role"),
    )

    objects = ScopedManager(organization="organization")

    class Meta:
        verbose_name = _("Organization membership")
        verbose_name_plural = _("Organization memberships")

    def __str__(self):
        return f"{self.user} in {self.organization} as {self.role}"
